/* ========================================================================== */
/*                                                                            */
/*   Radio.c                                                                  */
/*   (c) 2008 St                                                              */
/*                                                                            */
/*   Program for ATmega8 controlled FM tuner                                  */
/*   Xtall frequency: 32 kHz                                                  */
/* ========================================================================== */
#include "HWRadio.h"
#include <avr/eeprom.h>

uint16_t Time = 0;			// Global variable for Time (minutes from start of week) 
uint8_t Seconds = 0;		// Doublesecond counter (max. value = 29)
volatile uint8_t Mode = 1;	// 1=ON, 0=OFF (sleep)
uint8_t ErrPower = 0;		// Counter power line errors
uint8_t Key = 0;
uint8_t Stereo = 1;
uint8_t Mute = 0;
volatile uint8_t MuteChanged = 0;
volatile uint8_t StereoChanged = 0;
volatile uint8_t FreqChanged = 0;
volatile uint8_t WakeUp = 0;

unsigned int freq=9900;		// Actual tuned frequency
// Array in EEPROM for stored frequencies (Write your 6 favourites here) 
uint16_t ee_mem[6] __attribute__((section(".eeprom"))) = {9700,9800,9900,10000,10100,10200};
// Memory array for actions in EEPROM (With 5 steps example schedule ...)
uint16_t ee_act[200] __attribute__((section(".eeprom"))) = {1,9000,2,0,3,9100,4,1,5,2,0xFFFF};
uint16_t ActPointer = 0;
uint16_t MaxItem = 0;
uint16_t NextTime;
uint16_t NextAction;

// Fetch next time and action from schedule
void GetNextAction()
{
	NextTime=eeprom_read_word(ee_act + (2*ActPointer));
	if(NextTime == 0xFFFF) {
		ActPointer=0;
		NextTime=eeprom_read_word(ee_act);
	}
	NextAction=eeprom_read_word(ee_act + (2*ActPointer+1));
	ActPointer++;
}

// Get frequency from memory preset 
void GetPreset(uint8_t MemNo)
{
	uint16_t FromMemory;
	FromMemory=eeprom_read_word(ee_mem + MemNo);
	if((FromMemory <= 10800)&&(FromMemory >= 8750)) freq=FromMemory; else freq=8750;
}

// Test key pressed, number is returned on pressing button
uint8_t KeyPressed(void)
{
	uint8_t Pressed;

	Pressed = TestKbd();
	if(Pressed) {
		if(Pressed == Key) return 0;
		return (Key = Pressed);
	}
	else Key = 0; return 0;	
}

// Write frequency to display
void WriteFreq(int16_t frekv)
{
	char Tenth,Hunds,Megah,TenMega;
	unsigned int AuxFre;

	AuxFre=frekv;
	Tenth=(AuxFre%10)|0x30;
	AuxFre/=10;
	Hunds=(AuxFre%10)|0x30;
	AuxFre/=10;
	Megah=(AuxFre%10)|0x30;
	AuxFre/=10;
	TenMega=(AuxFre%10)|0x30;
	AuxFre/=10;
	if(AuxFre) WriteChar('1'); else WriteChar(' ');
	WriteChar(TenMega);
	WriteChar(Megah);
	WriteChar('.');
	WriteChar(Hunds);
	WriteChar(Tenth);
}

// Write time to display (Day and Time are alternating)
void WriteTime(int16_t time)
{
	unsigned char Day,Hours,Minutes;

	Minutes = time % 60;
	Hours = time / 60;
	Hours %= 24;

	if(Seconds & 0x01) {
		Day = time / 1440;
		switch(Day) {
			case 0: WriteText(" Mon "); break;
			case 1: WriteText(" Tue "); break;
			case 2: WriteText(" Wed "); break;
			case 3: WriteText(" Thu "); break;
			case 4: WriteText(" Fri "); break;
			case 5: WriteText(" Sat "); break;
			case 6: WriteText(" Sun "); break;
		}
		return;
	}

	Day = (Hours/10)|0x30;
	Hours %= 10; Hours |= 0x30;
	WriteChar(Day);WriteChar(Hours);
	WriteChar(':');
	Day = (Minutes / 10)|0x30;
	Minutes %= 10; Minutes |= 0x30;
	WriteChar(Day);WriteChar(Minutes);
}

// Start-up tuner
void StartTuner(int16_t frequency)
{
	POWER_ON;
  	_delay_loop_2(50000);
  	_delay_loop_2(50000);
  	_delay_loop_2(50000);
  	_delay_loop_2(50000);
	InitDisplay();
	SetDispPos(1,0);
	WriteFreq(frequency);
  	_delay_loop_2(500);
	SetFrequencyLM(frequency);
	uart_init();
}

// Auxiliary function for Menu item text on display
void MenuText(uint8_t item)
{
	switch(item) {
	case 0: WriteText("  Tune  "); break;
	case 1: WriteText(" Store  "); break;
	case 2: WriteText("Mono/Ste"); break;
	case 3: WriteText("  Mute  "); break;
	case 4: WriteText("Set time"); break;
	}
}

// Process menu item
void MenuItem(item)
{
	uint8_t Memory=0;
	uint16_t Day,Hr,Min;

	// Set new frequency
	if(item == 0) {
		SetDispPos(1,0);
		WriteFreq(freq);
		while(Key != KEYESC) {
		  	_delay_loop_2(2500);	// 10 msec delay
			if(KeyPressed()) {
				switch(Key) {
				case KEYUP: if(freq < 10800) freq+=5; break;
				case KEYDOWN: if(freq > 8750) freq-=5; break;
				}
				SetDispPos(1,0);
				WriteFreq(freq);
				SetFrequencyLM(freq);
			}
		}
	}
	// Store new frequency to memory preset
	if(item == 1) {
		SetDispPos(0,0);
		WriteText("Memory 1");
		while(Key != KEYESC) {
		  	_delay_loop_2(2500);	// 10 msec delay
			if(KeyPressed()) {
				switch(Key) {
				case KEYMEM1: Memory=0; break;
				case KEYMEM2: Memory=1; break;
				case KEYMEM3: Memory=2; break;
				case KEYMEM4: Memory=3; break;
				case KEYMEM5: Memory=4; break;
				case KEYMEM6: Memory=5; break;
				case KEYMENU: eeprom_write_word((uint16_t *)(Memory*2), freq); 
						SetDispPos(6,1);
						WriteChar('d');
					  	_delay_loop_2(50000);	// 200 msec delay
						SetDispPos(6,1);
						WriteChar(' ');break;
				}
				SetDispPos(0,0);
				WriteText("Memory ");
				WriteChar((Memory+1) | 0x30);
			}
		}
	}
	// Switch MONO/STEREO mode
	if(item == 2) {
		SetDispPos(1,0);
		if(Stereo) WriteText("Stereo");else WriteText(" Mono ");
		while(Key != KEYESC) {
		  	_delay_loop_2(2500);	// 10 msec delay
			if(KeyPressed()) {
				switch(Key) {
				case KEYUP: Stereo=1; STEREO_ON; break;
				case KEYDOWN: Stereo=0; STEREO_OFF; break;
				}
				SetDispPos(1,0);
				if(Stereo) WriteText("Stereo");else WriteText(" Mono ");
			}
		}
	}
	// Set MUTE/LOUD mode
	if(item == 3) {
		SetDispPos(0,0);
		if(Mute) WriteText("Mute ON ");else WriteText("Mute OFF");
		while(Key != KEYESC) {
		  	_delay_loop_2(2500);	// 10 msec delay
			if(KeyPressed()) {
				switch(Key) {
				case KEYUP: Mute=1; MUTE_ON; break;
				case KEYDOWN: Mute=0; MUTE_OFF; break;
				}
				SetDispPos(0,0);
				if(Mute) WriteText("Mute ON ");else WriteText("Mute OFF");
			}
		}
	}
	// Set new time in tuner
	if(item == 4) {
		cli();
		Day=Time/1440;
		Min=Time%60;
		Hr=Time/60-Day*24;
		sei();
		SetDispPos(0,0);
		WriteChar(Day | 0x30);
		WriteChar(' ');
		WriteChar((Hr/10) | 0x30);
		WriteChar((Hr%10) | 0x30);
		WriteChar(':');
		WriteChar((Min/10) | 0x30);
		WriteChar((Min%10) | 0x30);
		while(Key != KEYESC) {
		  	_delay_loop_2(2500);	// 10 msec delay
			if(KeyPressed()) {
				switch(Key) {
				case KEYUP: if(Day<6) Day++; else Day=0; break;
				case KEYDOWN: if(Day>0) Day--; else Day=6; break;
				case KEYHUP: if(Hr<23) Hr++; else Hr=0; break;
				case KEYHDOWN: if(Hr>0) Hr--; else Hr=23; break;
				case KEYMUP: if(Min<59) Min++; else Min=0; break;
				case KEYMDOWN: if(Min>0) Min--; else Min=59; break;
				case KEYMENU: cli(); Time=Day*1440+Hr*60+Min;sei();break;
				}
				SetDispPos(0,0);
				WriteChar(Day | 0x30);
				WriteChar(' ');
				WriteChar((Hr/10) | 0x30);
				WriteChar((Hr%10) | 0x30);
				WriteChar(':');
				WriteChar((Min/10) | 0x30);
				WriteChar((Min%10) | 0x30);
			}
		}
	}
}

// Function for menu item selection
void Menu(void)
{
	uint8_t item=0;

	ClrDisp();
	SetDispPos(2,0);
	WriteText("Menu");
	SetDispPos(0,1);
	MenuText(0);
	while(Key != KEYESC) {
	  	_delay_loop_2(2500);	// 10 msec delay
		if(KeyPressed()) {
			switch(Key) {
			case KEYUP: item++; if(item == 5) item=0; break;
			case KEYDOWN: if(item) item--; else item=4; break;
			case KEYMENU: MenuItem(item); break;
			}
			SetDispPos(0,1);
			MenuText(item);
		}
	}
	ClrDisp();
}

// Process character from serial line
void SerialChar(char SerChar)
{
	uint16_t i,auxword;
	char outch;

	switch(SerChar) {
	// No operation (for connection test)
	case 'N': uart_putchar('n'); break;
	// Set new next action
	case 'C': if(uart_getword(&auxword) && (auxword<100) && (auxword<=MaxItem)) {
				ActPointer=auxword; GetNextAction(); uart_putchar('c');}
				else uart_putchar('x'); break;
	// Decoders mode
	case 'D': outch='A';if(Stereo)outch+=1;if(Mute)outch+=2;uart_putchar(outch);uart_putchar('d'); break;
	// Set moNo output
	case 'O': StereoChanged=1; Stereo=0; uart_putchar('o'); break;
	// Set Stereo output
	case 'S': StereoChanged=1; Stereo=1; uart_putchar('s'); break;
	// Mute ON - stop output
	case 'M': MuteChanged=1; Mute=1; uart_putchar('m'); break;
	// Loud output - switch OFF mute
	case 'L': MuteChanged=1; Mute=0; uart_putchar('l'); break;
	// Set Time
	case 'T': if(uart_getword(&auxword) && (auxword<10080)) {
				cli();Time=auxword;sei(); uart_putchar('t');}
				else uart_putchar('x'); break;
	// Read Internal time
	case 'I': cli();auxword=Time;sei(); uart_putword(auxword);uart_putchar('i'); break;
	// Read Actual tuned frequency
	case 'A': uart_putword(freq); uart_putchar('a'); break;
	// Set Frequency - tune the radio
	case 'F': if(uart_getword(&auxword) && (auxword<=10800) && (auxword>=8750)) {
				freq=auxword; FreqChanged=1; uart_putchar('f');}
				else uart_putchar('x'); break;
	// Read memory Presets
	case 'P': for(i=0;i<6;i++) uart_putword(eeprom_read_word(ee_mem + i)); uart_putchar('p'); break;
	// Memory actions Read
	case 'R': for(i=0;i<200;i++) uart_putword(eeprom_read_word(ee_act + i)); uart_putchar('r'); break;
	// Write User memory presets
	case 'U': outch='u';i=0;
			  while((i<6)&&(outch=='u'))
				if(uart_getword(&auxword) && (auxword<=10800) && (auxword>=8750))
					{eeprom_write_word((uint16_t *)(i*2), auxword); i++;}
				else outch='x';
			  uart_putchar(outch); break;
	// Write User memory presets
	case 'W': outch='w';i=0;
			  while((i<100)&&(outch=='w')) {
				if(uart_getword(&auxword) && (auxword<=10080)) // Get and save next time
					eeprom_write_word(ee_act+i*2, auxword);
				else outch='v';
				if(uart_getword(&auxword) && (auxword<=10800)) // Get and save next action
					{eeprom_write_word(ee_act+i*2+1, auxword);i++;}
				else outch='v';
			  }
			  if(i>0) MaxItem=i-1; else MaxItem=0;
			  eeprom_write_word(ee_act+i*2,0xFFFF);
			  uart_putchar(outch); break;
	}
}
// ****************************
//    Main program function  //
// ****************************
int main(void) {
	char LastSec, inpchar;

	ASSR = 0x08;
	PORTB = 0x00;
	PORTC = 0xFF;
    DDRB = 0x3F;
    DDRD = 0xF3;
	PORTD = 0x0C;

	// Begin with switch OFF an ON 230V
	POWER_OFF;
  	_delay_loop_2(50000);
  	_delay_loop_2(50000);
  	_delay_loop_2(50000);
  	_delay_loop_2(50000);
  	_delay_loop_2(50000);
	StartTuner(freq);
	GetNextAction();
	/* Enable timer 1 overflow interrupt. */
	TCCR2 = 0x06;
	TIMSK = _BV (TOIE2);
	sei ();

/***** Main loop of program *****/
  for(;;) {
  	// Mode - Tuner is ON
  	if(Mode) {
		// Test power line status
		if(PowerLine()) ErrPower=0; else {
			ErrPower++;
			if(ErrPower == 100) {Mode = 0; ErrPower=0;}
		}
		// Set Mode, Mute, Frequency if changed
		if(WakeUp && Mode) {
			WakeUp = 0x00;
			StartTuner(freq);
		}
		// Was Mute changed?
		if(MuteChanged) {
			MuteChanged = 0x00;
			if(Mute) {MUTE_ON;SetDispPos(0,0);WriteChar('M');}
			else {MUTE_OFF;SetDispPos(0,0);WriteChar(' ');}
		}
		// Was Stereo changed?
		if(StereoChanged) {
			StereoChanged = 0x00;
			if(Stereo) STEREO_ON;
			else STEREO_OFF;
		}
		// Was frequency changed?
		if(FreqChanged) {
			FreqChanged = 0x00;
			SetFrequencyLM(freq);
			SetDispPos(1,0);
			WriteFreq(freq);
		}
		// Write time to display
		if(LastSec != Seconds) {
			SetDispPos(2,1);
			WriteTime(Time);
			LastSec = Seconds;
		}
		// Test keyboard
		if(KeyPressed()) {
			switch(Key) {
			case KEYMEM1: GetPreset(0); uart_putchar(0x31); break;
			case KEYMEM2: GetPreset(1); uart_putchar(0x32); break;
			case KEYMEM3: GetPreset(2); uart_putchar(0x33); break;
			case KEYMEM4: GetPreset(3); uart_putchar(0x34); break;
			case KEYMEM5: GetPreset(4); uart_putchar(0x35); break;
			case KEYMEM6: GetPreset(5); uart_putchar(0x36); break;
			case KEYESC: Mode = 0; break;
			case KEYMENU: Menu(); break;
			}
			SetFrequencyLM(freq);
			SetDispPos(0,0);
			if(Mute) WriteChar('M'); else WriteChar(' ');
			WriteFreq(freq);
		}
		// PC control through UART
		if(uart_getchar(&inpchar)) SerialChar(inpchar);
	  	_delay_loop_2(2500);	// 10 msec delay
	}
	// Switch OFF tuner - mode is OFF!
	if(!(Mode)) {
		POWER_OFF;
		UCSRB = 0;	/* stop tx/rx */
		MCUCR = 0xBA;
		asm ("sleep" ::);
		MCUCR = 0x0A;
	}
  } /* loop forever */
  /* please make sure that you never leave main */
  return (0);
}

/*
** ===================================================================
**     Interrupt handler
** ===================================================================
*/
ISR (TIMER2_OVF_vect)	// Every 2 seconds
{
	// Increment time variables
	Seconds++;
	if(Seconds == 30) {
		Seconds = 0;
		Time++;
		if(Time == 10080) Time = 0;
	}
	// Automatic action is allowed in 0 sec only
	if(!Seconds) {
		if(Time == NextTime) {
			switch (NextAction) {
			case 0: Mode = 0; break;
			case 1: Mute = 1; MuteChanged=1; break;
			case 2: Mute = 0; MuteChanged=1; break;
			default: if((NextAction <= 10800)&&(NextAction >= 8750)) {
				if(!(Mode)) {Mode=1; WakeUp=1;}
				freq = NextAction;
				FreqChanged=1;
				}
			}
			GetNextAction();
		}
	}
	// Manual wake-up?
	if(MenuKey() && (!(Mode & 0x01))) {
		WakeUp=1;
		Mode |= 0x01;
		Key = KEYMENU;
	}
}

/* END */
